#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""The ``gcp_utils`` module contains common functions used to interact with
GCP resources.
"""

import datetime
import os

from concurrent.futures import TimeoutError
from google.cloud import bigquery, pubsub_v1, storage
from google.cloud.logging_v2.logger import Logger
from google.cloud.pubsub_v1.subscriber.futures import StreamingPullFuture
from google.cloud.pubsub_v1.types import PubsubMessage, ReceivedMessage
import json
from typing import Callable, List, TYPE_CHECKING, Optional, Union

# load pandas only when necessary. it hogs memory on Cloud Functions.
if TYPE_CHECKING:
    import pandas as pd


# get project id from environment variable, else default to production project
# cloud functions use GCP_PROJECT
if "GCP_PROJECT" in os.environ:
    project_id_default = os.getenv("GCP_PROJECT")
else:
    project_id_default = os.getenv("GOOGLE_CLOUD_PROJECT", "ardent-cycling-243415")


# --- Pub/Sub --- #
def publish_pubsub(
    topic_name: str,
    message: Union[bytes, dict],
    project_id: Optional[str] = None,
    attrs: Optional[dict] = None,
    publisher: Optional[pubsub_v1.PublisherClient] = None
) -> str:
    """Publish messages to a Pub/Sub topic.

    Wrapper for `google.cloud.pubsub_v1.PublisherClient().publish()`.
    See also: https://cloud.google.com/pubsub/docs/publisher#publishing_messages.

    Args:
        topic_name: The Pub/Sub topic name for publishing alerts.

        message: The message to be published.

        project_id: GCP project ID for the project containing the topic.
                    If None, the module's `project_id_default` will be used.

        attrs: Message attributes to be published.

        publisher: An instantiated PublisherClient.
                   Use this kwarg if you are calling this function repeatedly.
                   The publisher will automatically batch the messages over a
                   small time window (currently 0.05 seconds) to avoid making
                   too many separate requests to the service.
                   This helps increase throughput. See
                   https://googleapis.dev/python/pubsub/1.7.0/publisher/index.html#batching

    Returns:
        published message ID
    """
    if project_id is None:
        project_id = project_id_default
    if publisher is None:
        publisher = pubsub_v1.PublisherClient()
    if attrs is None:
        attrs = {}

    # enforce bytes type for message
    if isinstance(message, dict):
        message = json.dumps(message).encode('utf-8')
    if not isinstance(message, bytes):
        raise TypeError('`message` must be bytes or a dict.')

    topic_path = publisher.topic_path(project_id, topic_name)

    future = publisher.publish(topic_path, data=message, **attrs)

    return future.result()


def pull_pubsub(
    subscription_name: str,
    max_messages: int = 1,
    project_id: Optional[str] = None,
    msg_only: bool = True,
    callback: Optional[Callable[[Union[ReceivedMessage, bytes]], bool]] = None,
    return_count: bool = False,
) -> Union[List[bytes], List[ReceivedMessage]]:
    """Pull and acknowledge a fixed number of messages from a Pub/Sub topic.

    Wrapper for the synchronous
    `google.cloud.pubsub_v1.SubscriberClient().pull()`.
    See also: https://cloud.google.com/pubsub/docs/pull#synchronous_pull

    Args:
        subscription_name: Name of the Pub/Sub subcription to pull from.

        max_messages: The maximum number of messages to pull.

        project_id: GCP project ID for the project containing the subscription.
                    If None, the module's `project_id_default` will be used.

        msg_only: Whether to work with and return the message contents only
                  or the full packet.
                  If `return_count` is True, it supersedes the returned object.

        callback: Function used to process each message.
                  Its input type is determined by the value of `msg_only`.
                  It should return True if the message should be acknowledged,
                  else False.

        return_count: Whether to return the messages or just the total number of
                      acknowledged messages.

    Returns:
        A list of messages
    """
    if project_id is None:
        project_id = project_id_default

    # setup for pull
    subscriber = pubsub_v1.SubscriberClient()
    subscription_path = subscriber.subscription_path(project_id, subscription_name)
    request = {
        "subscription": subscription_path,
        "max_messages": max_messages,
    }

    # wrap in 'with' block to automatically call close() when done
    with subscriber:

        # pull
        response = subscriber.pull(request)

        # unpack the messages
        message_list, ack_ids = [], []
        for received_message in response.received_messages:

            if msg_only:
                # extract the message bytes and append
                msg_bytes = received_message.message.data
                message_list.append(msg_bytes)
                # perform callback, if requested
                if callback is not None:
                    success = callback(msg_bytes)

            else:
                # append the full message
                message_list.append(received_message)
                # perform callback, if requested
                if callback is not None:
                    success = callback(received_message)

            # collect ack_id, if appropriate
            if (callback is None) or (success):
                ack_ids.append(received_message.ack_id)

        # acknowledge the messages so they will not be sent again
        if len(ack_ids) > 0:
            ack_request = {
                "subscription": subscription_path,
                "ack_ids": ack_ids,
            }
            subscriber.acknowledge(ack_request)

    if not return_count:
        return message_list
    else:
        return len(message_list)


def streamingPull_pubsub(
    subscription_name: str,
    callback: Callable[[PubsubMessage], None],
    project_id: str = None,
    timeout: int = 10,
    block: bool = True,
    flow_control: Optional[dict] = None,
) -> Union[None, StreamingPullFuture]:
    """Pull and process Pub/Sub messages continuously in a background thread.

    Wrapper for the asynchronous
    `google.cloud.pubsub_v1.SubscriberClient().subscribe()`.
    See also: https://cloud.google.com/pubsub/docs/pull#asynchronous-pull

    Args:
        subscription_name: The Pub/Sub subcription to pull from.

        callback: The callback function containing the message processing and
                  acknowledgement logic.

        project_id: GCP project ID for the project containing the subscription.
                    If None, the module's `project_id_default` will be used.

        timeout: The number of seconds before the `subscribe` call times out and
                 closes the connection.

        block: Whether to block while streaming messages or return the
               StreamingPullFuture object for the user to manage separately.

    Returns:
        If `block` is False, immediately returns the StreamingPullFuture object that
        manages the background thread.
        Call its `cancel()` method to stop streaming messages.
        If `block` is True, returns None once the streaming encounters an error
        or timeout.
    """
    if project_id is None:
        project_id = project_id_default

    if flow_control is None:
        flow_control = {}

    subscriber = pubsub_v1.SubscriberClient()
    subscription_path = subscriber.subscription_path(project_id, subscription_name)

    # start receiving and processing messages in a background thread
    streaming_pull_future = subscriber.subscribe(
        subscription_path, callback, flow_control=flow_control
    )

    if block:
        # block until timeout duration is reached or an error is encountered
        with subscriber:
            try:
                streaming_pull_future.result(timeout=timeout)
            except TimeoutError:
                streaming_pull_future.cancel()  # Trigger the shutdown.
                streaming_pull_future.result()  # Block until the shutdown is complete.

    else:
        return streaming_pull_future


def purge_subscription(subscription):
    """Purge all messages from the subscription."""
    client = pubsub_v1.SubscriberClient()
    sub = f"projects/{os.getenv('GOOGLE_CLOUD_PROJECT')}/subscriptions/{subscription}"
    _ = client.seek(request=dict(subscription=sub, time=datetime.datetime.now()))


# --- BigQuery --- #
def insert_rows_bigquery(
    table_id: str,
    rows: List[dict],
    project_id: Optional[str] = None,
):
    """Insert rows into a table using the streaming API.

    Args:
        table_id:   Identifier for the BigQuery table in the form
                    {dataset}.{table}. For example, 'ztf_alerts.alerts'.
        rows:       Data to load in to the table. Keys must include all required
                    fields in the schema. Keys which do not correspond to a
                    field in the schema are ignored.
        project_id: GCP project ID for the project containing the topic.
                    If None, the module's `project_id_default` will be used.
    """
    if project_id is None:
        project_id = project_id_default

    bq_client = bigquery.Client(project=project_id)
    table = bq_client.get_table(table_id)
    errors = bq_client.insert_rows(table, rows)
    return errors


def load_dataframe_bigquery(
    table_id: str,
    df: "pd.DataFrame",
    project_id: Optional[str] = None,
    use_table_schema: bool = True,
    logger: Optional[Logger] = None,
):
    """Load a dataframe to a table.

    Args:
        table_id: Identifier for the BigQuery table in the form
            {dataset}.{table}. For example, 'ztf_alerts.alerts'.
        df: Data to load in to the table. If the  dataframe schema does not match the
            BigQuery table schema, must pass a valid `schema`.
        project_id: GCP project ID for the project containing the topic.
                    If None, the module's `project_id_default` will be used.
        use_table_schema: Conform the dataframe to the table schema by converting
                          dtypes and dropping extra columns.
        logger: If not None, messages will be sent to the logger. Else, print them.
    """
    # setup
    if project_id is None:
        project_id = project_id_default
    bq_client = bigquery.Client(project=project_id)
    table = bq_client.get_table(table_id)

    if use_table_schema:
        my_df = df.reset_index()

        # set a job_config; bigquery will try to convert df.dtypes to match table schema
        job_config = bigquery.LoadJobConfig(schema=table.schema)

        # make sure the df has the correct columns
        bq_col_names = [s.name for s in table.schema]
        # pad missing columns
        missing = [c for c in bq_col_names if c not in my_df.columns]
        for col in missing:
            my_df[col] = None
        # drop extra columns
        dropped = list(set(my_df.columns) - set(bq_col_names))  # grab so we can report
        my_df = my_df[bq_col_names]
        # tell the user what happened
        if len(dropped) > 0:
            msg = f'Dropping columns not in the table schema: {dropped}'
            if logger is not None:
                logger.log_text(msg, severity='INFO')
            else:
                print(msg)

    else:
        my_df = df
        job_config = None

    # load the data
    job = bq_client.load_table_from_dataframe(my_df, table_id, job_config=job_config)
    job.result()  # Wait for the job to complete.

    # report the results
    msg = (
        f"Loaded {job.output_rows} rows to BigQuery table {table_id}.\n"
        f"The following errors were generated: {job.errors}"
    )
    if logger is not None:
        severity = 'DEBUG' if job.errors is not None else 'INFO'
        logger.log_text(msg, severity=severity)
    else:
        print(msg)


def query_bigquery(
    query: str,
    project_id: Optional[str] = None,
    job_config: Optional[bigquery.job.QueryJobConfig] = None,
) -> bigquery.job.QueryJob:
    """Query BigQuery.

    Args:
        query:
            SQL query statement.
        project_id:
            The GCP project id that will be used to make the API call.
            If not provided, the Pitt-Google production project id will be used.
        job_config:
            Optional job config to send with the query.

    Example query:
        ``
        query = (
            f'SELECT * '
            f'FROM `{dataset_project_id}.{dataset}.{table}` '
            f'WHERE objectId={objectId} '
        )
        ``

    Examples of working with the query_job:
        ``
        # Cast it to a DataFrame:
        query_job.to_dataframe()

        # Iterate row-by-row
        for r, row in enumerate(query_job):
            # row values can be accessed by field name or index
            print(f"objectId={row[0]}, candid={row['candid']}")
        ``
    """
    if project_id is None:
        project_id = project_id_default

    bq_client = bigquery.Client(project=project_id)
    query_job = bq_client.query(query, job_config=job_config)

    return query_job


# --- Cloud Storage --- #
def cs_download_file(localdir: str, bucket_id: str, filename: Optional[str] = None):
    """
    Args:
        localdir:   Path to local directory where file(s) will be downloaded to.
        bucket_id:  Name of the GCS bucket, not including the project ID.
                    For example, pass 'ztf-alert_avros' for the bucket
                    'ardent-cycling-243415-ztf-alert_avros'.
        filename:   Name or prefix of the file(s) in the bucket to download.
    """
    # connect to the bucket and get an iterator that finds blobs in the bucket
    storage_client = storage.Client(project_id_default)
    bucket_name = f'{project_id_default}-{bucket_id}'
    print(f'Connecting to bucket {bucket_name}')
    bucket = storage_client.get_bucket(bucket_name)
    blobs = storage_client.list_blobs(bucket, prefix=filename)  # iterator

    # download the files
    for blob in blobs:
        local_path = f'{localdir}/{blob.name}'
        blob.download_to_filename(local_path)
        print(f'Downloaded {local_path}')


def cs_upload_file(local_file: str, bucket_id: str, bucket_filename: Optional[str] = None):
    """
    Args:
        local_file: Path of the file to upload.
        bucket_id:  Name of the GCS bucket, not including the project ID.
                    For example, pass 'ztf-alert_avros' for the bucket
                    'ardent-cycling-243415-ztf-alert_avros'.
        bucket_filename:    String to name the file in the bucket. If None,
                            bucket_filename = local_filename.
    """
    if bucket_filename is None:
        bucket_filename = local_file.split('/')[-1]

    # connect to the bucket
    storage_client = storage.Client(project_id_default)
    bucket_name = f'{project_id_default}-{bucket_id}'
    print(f'Connecting to bucket {bucket_name}')
    bucket = storage_client.get_bucket(bucket_name)

    # upload
    blob = bucket.blob(bucket_filename)
    blob.upload_from_filename(local_file)
    print(f'Uploaded {local_file} as {bucket_filename}')



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Pitt-Google types."""

from dataclasses import dataclass
import logging
from typing import NamedTuple, Union

from .schema_maps import get_key, get_value


logger = logging.getLogger(__name__)


class _AlertIds(NamedTuple):
    """Simple class for IDs associated with an alert."""

    sourceId: Union[str, int, float, None] = None
    objectId: Union[str, int, float, None] = None


@dataclass
class AlertFilename:
    """Filename of an alert stored to a Cloud Storage bucket."""

    class ParsedFilename(NamedTuple):
        """Component parts of an ``AlertFilename``."""

        objectId: Union[str, int, float, None] = None
        sourceId: Union[str, int, float, None] = None
        topic: Union[str, None] = None
        format: str = "avro"

    def __init__(self, aname: Union[str, dict]):
        """Set self.name to the filename and self.parsed to a ``ParsedFilename``.

        Args
            aname:
                If str, full filename of the alert.
                If dict, components to create the filename. Required key/value pairs
                are those needed to create a ``ParsedFilename``.
                Extra keys are ignored.
        """
        if isinstance(aname, str):
            self.name = aname
            self.parsed = self.ParsedFilename._make(aname.split("."))

        elif isinstance(aname, dict):
            self.parsed = self.ParsedFilename(
                **dict(
                    (k, v) for k, v in aname.items() if k in self.ParsedFilename._fields
                )
            )
            self.name = ".".join(str(i) for i in self.parsed)


@dataclass
class AlertIds:
    """IDs associated with an alert."""

    def __init__(self, schema_map, **kwargs):
        """Initialize class instance.

        kwargs will be sent to `self.extract_ids()` and may contain keys:
            alert_dict, attrs, filename
        """
        self.schema_map = schema_map

        if any(key in kwargs for key in ["alert_dict", "attrs", "filename"]):
            self._ids = self.extract_ids(**kwargs)
        else:
            self._ids = None

        self._id_keys = None
        self._sourceId = None
        self._objectId = None

    @property
    def ids(self):
        """Alert IDs."""
        if self._ids is None:
            logger.warning(
                (
                    "No IDs have been set. Please call the extract_ids() method or "
                    "reinitialize with appropriate data."
                )
            )
        return self._ids

    def extract_ids(self, alert_dict=None, attrs=None, filename=None, **kwargs):
        """Extract IDs to an `_AlertIds` object and return it.

        Attempts to extract IDs from alert_dict, attrs, and filename, in that order.
        kwargs are ignored and only provided for the caller's convenience.
        """
        if alert_dict is not None:
            ids = _AlertIds(
                get_value("sourceId", alert_dict, self.schema_map),
                get_value("objectId", alert_dict, self.schema_map),
            )

        elif attrs is not None:
            id_keys = self.id_keys
            ids = _AlertIds(
                attrs.get(id_keys.sourceId), attrs.get(id_keys.objectId)
            )

        elif filename is not None:
            parsed = AlertFilename(filename).parsed
            ids = _AlertIds(parsed.sourceId, parsed.objectId)

        return ids

    @property
    def id_keys(self):
        """Return the ID key names used by the survey."""
        if self._id_keys is None:
            self._id_keys = _AlertIds(
                get_key("sourceId", self.schema_map),
                get_key("objectId", self.schema_map),
            )
        return self._id_keys

    @property
    def sourceId(self):
        """Return the sourceId."""
        if self._sourceId is None:
            self._sourceId = self.ids.sourceId
        return self._sourceId

    @property
    def objectId(self):
        """Return the objectId."""
        if self._objectId is None:
            self._objectId = self.ids.objectId
        return self._objectId



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
""" Simulate the consumer by publishing alerts to a Pub/Sub topic.
"""

import os
import sys
import time
from typing import Optional, Tuple, Union

from google.cloud import pubsub_v1


# get project id from environment variable, else default to production project
# cloud functions use GCP_PROJECT
if "GCP_PROJECT" in os.environ:
    PROJECT_ID = os.getenv("GCP_PROJECT")
else:
    PROJECT_ID = os.getenv("GOOGLE_CLOUD_PROJECT", "ardent-cycling-243415")


def publish_stream(
    alert_rate: Union[Tuple[int, str], str],
    instance: Optional[Tuple[str, str]] = None,
    runtime: Optional[Tuple[int, str]] = None,
    publish_batch_every: Tuple[int, str] = (5, 'sec'),
    sub_id: Optional[str] = None,
    topic_id: Optional[str] = None,
    nack: bool = False,
    auto_confirm: bool = False
):
    """Pulls messages from from a Pub/Sub subscription determined by either
    `instance` or `sub_id`, and publishes them to a topic determined by either
    `instance` or `topic_id`. Specific options are described in the docs.

    Args:
        alert_rate:  Desired rate at which alerts will be published.

        instance:   (survey, testid). Keywords of the broker instance. Used to
                    determine the subscription and topic. If `None`, `sub_id`
                    and `topic_id` must be valid names. If both `instance` and
                    `sub_id`/`topic_id` are passed, `sub_id`/`topic_id` will
                    prevail.

        runtime:    Desired length of time simulator runs for.

        publish_batch_every:    Simulator will sleep for this amount of time
                                between batches

        sub_id:     Name of the Pub/Sub subscription from which to pull alerts.
                    If `None`, `instance` must contain valid keywords, and then
                    the production instance reservoir
                    '{survey}-alerts-reservoir' will be used.

        topic_id:   Name of the Pub/Sub topic to which alerts will be published.
                    If `None`, `instance` must contain valid keywords, and then
                    the topic '{survey}-alerts-{testid}' will be used.

        nack:       Whether to "nack" (not acknowledge) the messages. If `True`,
                    messages are published to the topic, but they are not
                    dropped from the subscription and so will be delivered again
                    at an arbitrary time in the future.

        auto_confirm:   Whether to automatically answer "Y" to the confirmation prompt.
    """

    pbeN, pbeU = publish_batch_every  # shorthand

    # get number of alerts to publish per batch
    alerts_per_batch, aRate_tuple = _get_number_alerts_per_batch(alert_rate, publish_batch_every)  # int, tuple

    # get number of batches to run
    Nbatches = _convert_runtime_to_Nbatches(runtime, publish_batch_every, aRate_tuple[1])

    # tell the user about the rates
    print(f"\nReceived desired alert_rate={aRate_tuple}, runtime={runtime}.")
    print(
        f"\nPublishing:\n\t{Nbatches} batches\n\teach with {alerts_per_batch} alerts\n\tat a rate of 1 batch per {pbeN} {pbeU} (plus processing time)\n\tfor a total of {Nbatches * alerts_per_batch} alerts")

    # publish the stream
    _do_publish_stream(instance, alerts_per_batch, Nbatches, publish_batch_every, sub_id, topic_id, nack, auto_confirm)


def _do_publish_stream(
    instance,
    alerts_per_batch,
    Nbatches,
    publish_batch_every,
    sub_id=None,
    topic_id=None,
    nack=False,
    auto_confirm=False
):
    # check units
    if publish_batch_every[1] != 'sec':
        raise ValueError("Units of publish_batch_every must = 'sec'.")

    # setup for subscription pulls and publishing
    subscriber, sub_path, request = _setup_subscribe(alerts_per_batch, instance, sub_id)
    publisher, topic_path = _setup_publish(alerts_per_batch, instance, topic_id)
    print(f"\nThis will\n\tPull from subscription: {sub_path}")
    print(f"and\n\tPublish to topic: {topic_path}\n")

    # make the user confirm
    _user_confirm(auto_confirm)
    print(f"\nPublishing...")

    b = 0
    while b < Nbatches:
        # get alerts from reservoir
        # response = subscriber.pull(request=request)  # for pubsub2+
        response = subscriber.pull(request)

        # publish alerts to topic, raise exception on failure
        _publish_received_messages(publisher, topic_path, response)

        # ack or nack subscription messages
        ack_ids = [msg.ack_id for msg in response.received_messages]
        _handle_acks(subscriber, sub_path, ack_ids, nack)

        # increment and sleep between batches
        b = b + 1
        time.sleep(publish_batch_every[0])


def _user_confirm(auto_confirm=False):
    if not auto_confirm:
        cont = input('Continue? [Y/n]: ') or 'Y'
        if cont not in ['Y', 'y']:
            sys.exit('Exiting consumer simulator.')


def _setup_subscribe(alerts_per_batch, instance=None, sub_id=None):
    if (instance is None) and (sub_id is None):
        raise ValueError('Must provide either `instance` or `sub_id`.')

    if sub_id is None:
        # use the reservoir of the production instance for this survey
        survey = instance[0]
        sub_id = f'{survey}-alerts-reservoir'

    subscriber = pubsub_v1.SubscriberClient()
    sub_path = subscriber.subscription_path(PROJECT_ID, sub_id)
    request = {
        "subscription": sub_path,
        "max_messages": alerts_per_batch,
    }

    return (subscriber, sub_path, request)


def _setup_publish(alerts_per_batch, instance=None, topic_id=None):
    if (instance is None) and (topic_id is None):
        raise ValueError('Must provide either `instance` or `topic_id`.')

    if topic_id is None:
        survey, testid = instance
        topic_id = f'{survey}-alerts-{testid}'

    # calls to publish are batched automatically
    # let's try to get all alerts into 1 publisher batch
    batch_settings = pubsub_v1.types.BatchSettings(max_messages=alerts_per_batch)
    # some default batch settings to be aware of:
    # max_messages = 100
    # max_bytes = 1 MB
    # max_latency = 10 ms
    publisher = pubsub_v1.PublisherClient(batch_settings)
    topic_path = publisher.topic_path(PROJECT_ID, topic_id)

    return (publisher, topic_path)


def _publish_received_messages(publisher, topic_path, sub_response):
    """ Iterate through messages and publish them, along with their attributes, to the Pub/Sub topic.
    """
    for msg in sub_response.received_messages:
        # print("Received message:", msg.message.data)
        attrs = msg.message.attributes  # pass msg attributes through

        future = publisher.publish(topic_path, msg.message.data, **attrs)  # non blocking

        future.add_done_callback(_callback)  # check for errors in a separate thread


def _handle_acks(subscriber, sub_path, ack_ids=[], nack=False):
    """ If nack is False, acknowledge messages, else nack them so they stay in the reservoir.
    """
    if not nack:
        request = {
            "subscription": sub_path,
            "ack_ids": ack_ids,
        }
        subscriber.acknowledge(request)
    else:
        request = {
            "subscription": sub_path,
            "ack_ids": ack_ids,
            "ack_deadline_seconds": 0,
        }
        subscriber.modify_ack_deadline(request)


def _callback(future):
    # Publishing failures are automatically retried, except for errors that do not warrant retries.
    message_id = future.result()  # raises an exception if publish ultimately failed
    # print(message_id)


def _convert_runtime_to_Nbatches(runtime, publish_batch_every, aRate_unit):
    if aRate_unit == 'once':
        Nbatches = 1

    elif type(runtime) != tuple:
        msg = "runtime must be given as a tuple, unless the alert_rate units are set to 'once'"
        raise ValueError(msg)

    else:
        pbeN, pbeU = publish_batch_every  # shorthands
        publish_runtime = _convert_time_to_publish_unit(runtime, pbeU)  # int
        Nbatches = _convert_publish_runtime_to_Nbatches(publish_runtime, pbeN)  # int

    return Nbatches


def _convert_publish_runtime_to_Nbatches(publish_runtime, publish_batch_every_N):
    """
    Args:
        publish_runtime (int or float): runtime requested by user, in units used by the publisher
        publish_batch_every_N (int): publisher will use a rate of 1 batch per publish_batch_every_N.
        Should be in the same units as publish_runtime.
    Returns:
        Nbatches (int): number of batches the publisher should complete, rounded to an int
    """
    Nbatches = publish_runtime * (1 / publish_batch_every_N)
    return int(Nbatches)


def _convert_time_to_publish_unit(runtime, publish_unit):
    """
    Args:
        runtime (int, str): (number, unit) to convert to publish_unit
        publish_unit (str): time units consumer will use for publish rate

    Returns:
        publish_runtime (int): runtime converted to publish_unit
    """

    rtN, rtU = runtime

    # convert the time to (N, 'sec')
    if publish_unit == 'sec':
        if rtU == 'sec':
            publish_runtime = rtN
        elif rtU == 'min':
            publish_runtime = rtN * 60
        elif rtU == 'hr':
            publish_runtime = rtN * 3600
        elif rtU == 'night':
            publish_runtime = rtN * 10 * 3600

    else:
        msg = f"received publish_units='{publish_unit}', but only configured for publish_unit='sec'"
        raise ValueError(msg)

    return publish_runtime


def _get_number_alerts_per_batch(alert_rate, publish_batch_every):
    """ Find the number of alerts to publish per batch
    """

    aRate_tuple = _convert_rate_to_tuple(alert_rate)  # tuple

    # if user requested 1 batch, return their number
    if aRate_tuple[1] == 'once':
        alerts_per_batch = aRate_tuple[0]

    # else convert to batch rate
    else:
        pbeN, pbeU = publish_batch_every
        avg_publish_rate = _convert_rate_to_publish_unit(aRate_tuple, publish_unit=pbeU)  # float
        alerts_per_batch = _get_number_alerts_per_batch_from_avg(avg_publish_rate, pbeN)  # int, per batch

    return (alerts_per_batch, aRate_tuple)


def _get_number_alerts_per_batch_from_avg(avg_publish_rate, publish_interval):
    return int(avg_publish_rate * publish_interval)


def _convert_rate_to_publish_unit(alert_rate, publish_unit='sec'):
    """
    Args:
        alert_rate (int, str): (number, unit) to convert to publish_unit
        publish_unit (str): time units consumer will use for publish rate

    Returns:
        alerts_per_unit (float): alert_rate in units per publish_unit
    """

    arN, arU = alert_rate

    # convert the rate to (N, 'perSec')
    if publish_unit == 'sec':
        if arU == 'perSec':
            alerts_per_unit = arN
        if arU == 'perMin':
            alerts_per_unit = arN / 60
        if arU == 'perHr':
            alerts_per_unit = arN / 3600
        if arU == 'perNight':
            alerts_per_unit = arN / 10 / 3600

    else:
        msg = f"received publish_units='{publish_unit}', but only configured for publish_unit='sec'"
        raise ValueError(msg)

    return alerts_per_unit


def _convert_rate_to_tuple(alert_rate):
    if type(alert_rate) == tuple:
        # type == tuple, so just return it
        aRate = alert_rate

    elif type(alert_rate) == str:
        aRate = _convert_rate_string_to_tuple(alert_rate)
        # returns tuple or raises ValueError

    else:
        msg = 'alert_rate must be a tuple or a string'
        raise ValueError(msg)

    return aRate


def _convert_rate_string_to_tuple(alert_rate):
    # convert strings to tuples

    if alert_rate == 'ztf-active-avg':
        aRate = (300000, 'perNight')

    elif alert_rate == 'ztf-live-max':
        aRate = (200, 'perSec')

    else:
        msg = f"'ztf-active-avg' and 'ztf-live-max' are the only strings currently configured for the alert_rate"
        raise ValueError(msg)

    return aRate



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""Classes and functions used for Pitt-Google broker testing."""

import base64
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
import logging
import os
from pathlib import Path
from typing import NamedTuple, Union

import numpy as np

from .gcp_utils import pull_pubsub
from .data_utils import load_alert, decode_alert
from .types import _AlertIds, AlertIds


logger = logging.getLogger(__name__)


@dataclass
class AlertPaths:
    """Paths to alerts stored locally."""

    def __init__(self, alert_dir: Union[str, Path] = "ztf"):
        """Load path to alerts directory and initialize variables.

        Args:
            alert_dir:
                Either the path to a directory containing alerts or one of the
                following strings.
                "ALERT_DIR" or "alert_dir": The path will be obtained from the
                    environment variable ALERT_DIR.
                "ztf" or "elasticc": The path will be obtained from the environment
                    variable given by ``f"ALERT_DIR_{alert_dir.upper()}"``
        """
        if alert_dir in ["ALERT_DIR", "alert_dir"]:
            self.alert_dir = Path(os.getenv("ALERT_DIR"))
        elif alert_dir in ["ztf", "elasticc"]:
            self.alert_dir = Path(os.getenv(f"ALERT_DIR_{alert_dir.upper()}"))
        else:
            self.alert_dir = Path(alert_dir)

        self.__mygen = None

    def gen(self):
        """Return a new generator of paths to Avro files in self.alert_dir."""
        return (p for p in self.alert_dir.glob("**/*.avro"))

    @property
    def path(self):
        """Return a path to an alert Avro file.

        This can be used to obtain a random path to an alert, or within a loop to
        continuously iterate over paths in the alert directory.

        The returned path is the next one in self._mygen (a path generator maintained
        internally).
        If you want more control (e.g., to ensure that you iterate over all files in
        the directory and/or do not recieve duplicate paths), use ``AlertPaths().gen()``
        instead.
        """
        return next(self._mygen)

    @property
    def _mygen(self):
        """Path generator used by self."""
        if self.__mygen is None:
            self.__mygen = self.gen()
        return self.__mygen


@dataclass
class Mock:
    """Mock data and objects useful for testing the broker."""

    def __init__(self, **kwargs):
        """Initialize attributes.

        kwargs may contain keys:
            modules:
                List of module names to mock.
            test_alert:
                An instance of TestAlert. This will be used to generate appropriate
                mock data (e.g., Pub/Sub message attributes containing the TestAlert's
                alert IDs).
            args and kwargs for TestAlert:
                These will be used to create a new instance of TestAlert.
        """
        self.kwargs = kwargs

        self._modules = kwargs.get("modules")
        self._module_results = None

        self._my_test_alert = kwargs.get("test_alert")
        self._id_tuples = None

        self._attrs = None

        self._cfinput = None

    # modules to mock
    @property
    def modules(self):
        """List of modules for which mock results should be created.

        If this is set manually, self._module_results will be recreated to accommodate.
        """
        return self._modules

    @modules.setter
    def modules(self, value):
        self._modules = value
        # force recreation of other properties
        self._module_results = None

    @modules.deleter
    def modules(self):
        self._modules = None
        # force recreation of other properties
        self._module_results = None

    # mocked module results
    @property
    def module_results(self):
        """Mock results for the pipeline module(s) given by self.modules."""
        if self._module_results is None:
            if self.modules is not None:
                self._module_results = self._generate_module_results()

        return self._module_results

    @module_results.setter
    def module_results(self, value):
        self._module_results = value

    @module_results.deleter
    def module_results(self):
        self._module_results = None

    def _generate_module_results(self):
        results = dict()

        if "SuperNNova" in self.modules:
            results["SuperNNova"] = self._supernnova_results

        self._module_results = results
        return self._module_results

    @property
    def _supernnova_results(self):
        """Return mocked results for SuperNNova module."""
        prob_class1 = np.random.uniform()
        return dict(
            [
                (self.id_keys.objectId, self.ids.objectId),
                (self.id_keys.sourceId, self.ids.sourceId),
                ("prob_class0", 1 - prob_class1),
                ("prob_class1", prob_class1),
                ("predicted_class", round(prob_class1)),
            ]
        )

    @property
    def my_test_alert(self):
        """Instance of TestAlert used by self to generate appropriate data."""
        if self._my_test_alert is None:
            kwargs = dict(self.kwargs)
            path = kwargs.pop("path", AlertPaths().path)
            schema_map = kwargs.pop("schema_map")
            self._my_test_alert = TestAlert(path, schema_map, **kwargs)
        return self._my_test_alert

    @my_test_alert.setter
    def my_test_alert(self, value):
        self._my_test_alert = value
        # force recreation of other properties
        self._id_tuples = None
        self._cfinput = None

    @my_test_alert.deleter
    def my_test_alert(self):
        self._my_test_alert = None
        # force recreation of other properties
        self._id_tuples = None
        self._cfinput = None

    # tuples of mocked alert ID keys and IDs
    @property
    def id_tuples(self):
        """Tuples of alert ID keys and IDs."""
        if self._id_tuples is None:

            if self._my_test_alert is None:
                sid = ("sourceId", f"mocksrc{int(1e9 * np.random.uniform())}")
                oid = ("objectId", f"mockobj{int(1e9 * np.random.uniform())}")

            else:
                sid = (
                    self._my_test_alert.id_keys.sourceId,
                    str(self._my_test_alert.ids.sourceId),
                )
                oid = (
                    self._my_test_alert.id_keys.objectId,
                    str(self._my_test_alert.ids.objectId),
                )

            self._id_tuples = (sid, oid)

        return self._id_tuples

    @id_tuples.setter
    def id_tuples(self, value):
        self._id_tuples = value

    @id_tuples.deleter
    def id_tuples(self):
        self._id_tuples = None

    # mocked attributes
    @property
    def attrs(self):
        """Mock attributes (dict) for a Pub/Sub message."""
        if self._attrs is None:
            sid, oid = self.id_tuples
            # create int POSIX timestamps in microseconds
            now = datetime.now(timezone.utc)
            tkafka = int(now.timestamp() * 1e6)
            # to convert back: datetime.fromtimestamp(tkafka / 1e6)
            tingest = int((now + timedelta(seconds=0.345678)).timestamp() * 1e6)

            self._attrs = dict(
                [
                    sid,
                    oid,
                    ("kafka.topic", "mock_topic_pittgoogle_test"),
                    ("kafka.timestamp", str(tkafka)),
                    ("ingest.timestamp", str(tingest)),
                ]
            )
        return self._attrs

    @attrs.setter
    def attrs(self, value):
        self._attrs = value

    @attrs.deleter
    def attrs(self):
        self._attrs = None

    # mocked Cloud Functions input
    class _CFInput(NamedTuple):
        """Alert data for input to a Cloud Function."""

        msg: dict
        context: NamedTuple

    class _CFContext(NamedTuple):
        """Mock context input for a Cloud Function."""

        event_id = int(1e12 * np.random.uniform())
        timestamp = (
            datetime.now(timezone.utc)
            .isoformat(timespec="milliseconds")
            .replace("+00:00", "Z")
        )
        # event_type = "google.pubsub.topic.publish"
        event_type = "broker_utils.testing.Mock"
        resource = "mock_resource"

    @property
    def cfinput(self):
        """Mock input to a Cloud Function."""
        if self._cfinput is None:

            # return immediately if no schema map was provided
            if self.kwargs.get("schema_map") is None:
                logger.error("A schema map is required to mock Cloud Function input.")
                return

            else:
                self._cfinput = self._CFInput(
                    msg=dict(
                        data=base64.b64encode(self.my_test_alert.msg_payload),
                        attributes=self.my_test_alert.mock.attrs,
                    ),
                    context=self._CFContext(),
                )

        return self._cfinput

    @cfinput.setter
    def cfinput(self, value):
        self._cfinput = value

    @cfinput.deleter
    def cfinput(self):
        self._cfinput = None


@dataclass
class TestAlert:
    """An alert packet and related functions useful for testing the broker."""

    def __init__(self, path, schema_map, **kwargs):
        """Load the alert from `path` and initialize attributes.

        kwargs may contain keys:
            drop_cutouts (bool):
                Wether to drop the image cutouts from the alert.
            serialize (str):
                "avro" or "json". Serialization format of the Pub/Sub message.
            mock (Mock):
                An instance of Mock containing mocked data which should be attached to
                the message.
            mock_modules (List[str]):
                List of modules for which data should be mocked. This will be passed to
                Mock() as keyword argument "modules". This is ignored if an instance of
                Mock is passed using the "mock" keyword argument.
        """
        self.path = path

        abytes = load_alert(path, "bytes")
        adict = decode_alert(
            abytes, drop_cutouts=kwargs.get("drop_cutouts", True), schema_map=schema_map
        )
        self.data = dict(bytes=abytes, dict=adict)

        aids = AlertIds(schema_map, alert_dict=self.data["dict"])
        self.id_keys = aids.id_keys
        self.ids = aids.ids

        self._serialize = kwargs.get("serialize", "json")
        self._msg_payload = None

        self._mock_modules = kwargs.get("mock_modules", None)
        self._mock = kwargs.get("mock", None)

    # message payload
    @property
    def msg_payload(self):
        """Pub/Sub message payload containing the alert and any mocked results."""
        if self._msg_payload is None:
            # create the message payload
            # alert data type appropriate for the format given by self.serialize
            dtype = {"json": "dict", "avro": "bytes"}
            alert = self.data[dtype[self.serialize]]

            if self.serialize == "avro":
                self._msg_payload = alert

                # currently can't publish an Avro serialized message with mock results
                if self.mock_modules is not None:
                    logger.warning(
                        (
                            "The published message will be an Avro serialized alert. "
                            "Mocked results will NOT be attached."
                        )
                    )

            else:
                if self.mock_modules is None:
                    self._msg_payload = alert

                else:
                    self._msg_payload = dict(alert=alert, **self.mock.module_results)

        return self._msg_payload

    # message serialization format
    @property
    def serialize(self):
        """One of "json" or "avro". Determines the format of the published message.

        If this is set manually, self._msg_payload will be recreated to accommodate.
        If this is manually deleted, it will be set to the default ("json") and
        self._msg_payload will be recreated to accommodate.
        """
        return self._serialize

    @serialize.setter
    def serialize(self, value):
        self._serialize = value
        # force recreation of other properties
        self._msg_payload = None

    @serialize.deleter
    def serialize(self):
        self._serialize = "json"
        # force recreation of other properties
        self._msg_payload = None

    @staticmethod
    def guess_serializer(topic):
        """Use the topic to kguess the format that the message is expected to be in."""
        avro_topics = ["alerts"]  # all others are json

        try:
            topic_name_stub = topic.split("-")[1]
        except IndexError:
            topic_name_stub = topic

        serialize = "avro" if topic_name_stub in avro_topics else "json"
        return serialize

    @property
    def mock_modules(self):
        """List of modules for which results should be mocked.

        If this is set manually, self._mock will be recreated to accommodate.
        """
        return self._mock_modules

    @mock_modules.setter
    def mock_modules(self, value):
        self._mock_modules = value
        # force recreation of other properties
        self._mock = None

    @mock_modules.deleter
    def mock_modules(self):
        self._mock_modules = None
        # force recreation of other properties
        self._mock = None

    @property
    def mock(self):
        """Mock data.

        If this is set manually, this will also set self.mock_modules = mock.modules
        """
        if self._mock is None:
            self._mock = Mock(modules=self.mock_modules, test_alert=self)
        return self._mock

    @mock.setter
    def mock(self, value):
        self._mock = value
        # set dependent attributes
        self.mock_modules = value.modules

    @mock.deleter
    def mock(self):
        self._mock = None


@dataclass
class IntegrationTestValidator:
    """Functions to validate an integration test."""

    def __init__(self, subscrip, published_alert_ids, schema_map, **kwargs):
        """Initialize attributes.

        kwargs can include keys:
            max_pulls (int):
                maximum number of times to pull the Pub/Sub subscription
                before quitting.
        """
        self.subscrip = subscrip
        self.max_pulls = kwargs.get("max_pulls", 4)
        self.pulled_msg_ids = None
        self.published_alert_ids = published_alert_ids
        self.unmatched_ids = None
        self.alert_ids = AlertIds(schema_map)  # load once, use to extract all msg IDs

    def run(self):
        """Pull the subscription and validate that ids match the published alerts."""
        self.pulled_msg_ids = self._pull()
        self.success, self.unmatched_ids = self._compare_ids()
        return self.success

    def _pull(self):
        pulled_msg_ids, i = [], 0
        while len(pulled_msg_ids) < len(self.published_alert_ids):
            max_messages = len(self.published_alert_ids) - len(pulled_msg_ids)
            msgs = pull_pubsub(self.subscrip, max_messages=max_messages, msg_only=False)
            if len(msgs) > 0:
                pulled_msg_ids += self._extract_ids(msgs)

            i += 1
            if i >= self.max_pulls:
                break

        return pulled_msg_ids

    def _extract_ids(self, msgs):
        """Extract the alert IDs from msgs.

        This method first guesses whether the IDs should be extracted from message
        attributes or alert filenames by checking the objectId attribute of the first
        message in msgs.
        Background: Messages generated by a GCS bucket contain an attribute called
        objectId which contains the filename of the object (i.e., file) in the bucket
        that triggered the message. This is a name collision with the alert's DIA
        objectId, which broker pipeline modules place in an attribute of the same name.
        """
        oid0 = msgs[0].message.attributes["objectId"]
        oid_is_filename = oid0.split(".")[-1] == "avro"

        if oid_is_filename:
            pulled_msg_ids = [
                self.alert_ids.extract_ids(filename=msg.message.attributes["objectId"])
                for msg in msgs
            ]

        else:
            pulled_msg_ids = [
                self.alert_ids.extract_ids(attrs=msg.message.attributes) for msg in msgs
            ]

        return pulled_msg_ids

    def _compare_ids(self):
        idsout = set(self.pulled_msg_ids)
        # self._pull() gets the ids from the message attributes or filenames,
        # which are always strings.
        # convert the published_alert_ids to the same type.
        idsin = set(
            [_AlertIds(*[str(id) for id in ids]) for ids in self.published_alert_ids]
        )

        # compare ID sets
        unmatched_ids = idsout.symmetric_difference(idsin)
        success = len(unmatched_ids) == 0

        # log results
        tmp = f"The message IDs pulled from subscription {self.subscrip}"
        if success:
            logger.info(f"Success! {tmp} match the input.")
        else:
            logger.debug(f"Something went wrong. {tmp} do not match the input.")

        # warn if idsout contains IDs that were not published
        if len(idsout.difference(idsin)) > 0:
            logger.warning(
                (
                    "Some IDs were pulled that were not reported as published. "
                    "You may want to purge the subcription using\n"
                    f"\tbroker_utils.gcp_utils.purge_subscription({self.subscrip})\n"
                    "and run the test again."
                )
            )

        return success, unmatched_ids



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""The ``schema_maps`` module loads and returns schema maps stored in
Cloud Storage.
"""

import os
from pathlib import Path
from typing import Optional, Union

from google.cloud import storage
from google.cloud.storage.blob import Blob
import yaml


# get project id from environment variable, else default to production project
# cloud functions use GCP_PROJECT
if "GCP_PROJECT" in os.environ:
    PROJECT_ID = os.getenv("GCP_PROJECT")
else:
    PROJECT_ID = os.getenv("GOOGLE_CLOUD_PROJECT", "ardent-cycling-243415")


def load_schema_map(
    survey: str,
    testid: str,
    schema: Optional[Union[Path, str]] = None,
) -> dict:
    """Return a dict mapping the survey's field names to an internal broker standard.

    Args:
        survey: Name of the survey associated with the broker instance.
                If `schema` is a `pathlib.Path`, `survey` is ignored
                (left as an arg for backward compatibility).
                Otherwise, it determines (in conjunction with `testid`) which GCS bucket
                the schema map will be loaded from. If `schema` is `None`,
                this will also determine which survey's schema map is returned.
        testid: If `schema` is a `pathlib.Path`, `testid` is ignored
                (left as an arg for backward compatibility).
                Otherwise, it determines (in conjunction with `survey`) which GCS bucket
                the schema map will be loaded from.
                If this is a bool, it must be False (indicating a production instance).
        schema: If this is a `pathlib.Path`: path to a local yaml file to be loaded.
                In this case, `survey` and `testid` are ignored.
                If this is a `str`: name of the survey who's schema map
                is to be loaded. It will be retrieved from a Cloud Storage bucket.
                If this is `None`, the schema map corresponding to `survey` will be
                will be retrieved from a Cloud Storage bucket.

    Returns:
        Dictionary mapping the survey's field names to an internal broker standard.
    """
    if not schema:
        # set schema to the survey name, map will be retrieved from a bucket
        schema = survey  # str

    if isinstance(schema, str):
        # download from storage bucket
        return _download_schema_map(survey, testid, schema)

    # else load from local file
    return load_yaml(schema)


def _download_schema_map(survey: str, testid: str, schema: str) -> dict:
    """Download the schema map from Cloud Storage and return as a dict.

    Args:
        survey: Name of the survey associated with the broker instance.
                Determines (in conjunction with `testid`) which GCS bucket
                the schema map will be loaded from. If `schema` is `None`,
                this will also determine which survey's schema map is returned.
        testid: Determines (in conjunction with `survey`) which GCS bucket
                the schema map will be loaded from.
                If this is a bool, it must be False (indicating a production instance).
        schema: Name of the survey who's schema map
                is to be loaded. It will be retrieved from a Cloud Storage bucket.

    Returns:
        Dictionary mapping the survey's field names to an internal broker standard.
    """
    # some setup
    bucket = _broker_bucket_name(survey, testid)
    schema = _schema_object_name(schema)
    client = storage.Client()

    # try to get the object from storage. blob = `None` if unsuccessful
    blob = client.bucket(bucket).get_blob(schema)  # 403 error (Forbidden) if not auth'd

    if not blob:
        # raise an error
        _ = client.get_bucket(bucket)  # 404 error (NotFound) if bucket doesn't exist
        raise ValueError(f"Bucket {bucket} exists but the object {schema} does not.")

    return load_yaml(blob)


def _broker_bucket_name(survey, testid):
    if testid in ["False", False]:
        return f'{PROJECT_ID}-{survey}-broker_files'
    else:
        return f'{PROJECT_ID}-{survey}-broker_files-{testid}'


def _schema_object_name(survey):
    return f'schema_maps/{survey}.yaml'


def load_yaml(fin: Union[Path, str, Blob]) -> dict:
    """Load a yaml file and return as a dict.

    Args:
        fin: Path-like or `google.cloud.storage.blob.Blob` object. Assumes yaml format.

    Returns:
        Dictionary mapping the survey's field names to an internal broker standard.
    """
    if isinstance(fin, Blob):
        with fin.open("rt") as f:
            return yaml.safe_load(f)  # dict

    with open(fin, "rb") as f:
        return yaml.safe_load(f)  # dict


def get_value(key: str, alert_dict: dict, schema_map: dict) -> Union[int, str]:
    """Return the alert_dict value corresponding to schema_map and key."""
    fullkey = schema_map.get(key)

    if isinstance(fullkey, str):
        return alert_dict.get(fullkey)

    elif isinstance(fullkey, list):
        if len(fullkey) == 1:
            return alert_dict.get(fullkey[0])

        elif len(fullkey) == 2:
            return alert_dict.get(fullkey[0]).get(fullkey[1])

        elif len(fullkey) == 3:
            return alert_dict.get(fullkey[0]).get(fullkey[1]).get(fullkey[2])


def get_key(key: str, schema_map: dict) -> Union[int, str]:
    """Return the key or final element in list corresponding to schema_map and key."""
    fullkey = schema_map.get(key)

    if isinstance(fullkey, str):
        return fullkey

    elif isinstance(fullkey, list):
        return fullkey[-1]



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""The ``data_utils`` module contains common functions used to manipulate
survey and broker data.
"""

from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Union

import fastavro
import json

# load pandas only when necessary. it hogs memory on Cloud Functions.
if TYPE_CHECKING:
    import pandas as pd

from .types import AlertIds


def load_alert(
    fin: Union[str, Path],
    return_as: str = "dict",
    **kwargs
) -> Union[bytes, dict, "pd.DataFrame"]:
    """Load alert from file at ``fin`` and return in format ``return_as``.

    Args:
        fin:        Path to an alert avro file.
        return_as:  Format the alert will be returned in. One of 'bytes' or argument
                    accepted by ``decode_alert``.
        kwargs:     Keyword arguments for ``decode_alert``.
    """
    if return_as == "bytes":
        with open(fin, "rb") as f:
            alert = f.read()
    else:
        alert = decode_alert(fin, return_as=return_as, **kwargs)
    return alert


def decode_alert(
    alert_avro: Union[str, Path, bytes],
    return_as: str = 'dict',
    drop_cutouts: bool = False,
    **kwargs
) -> Union[dict, "pd.DataFrame"]:
    """Load an alert Avro and return in requested format.

    Wraps `alert_avro_to_dict()` and `alert_dict_to_dataframe()`.

    Args:
        alert_avro:   Either the path of Avro file to load, or
            the bytes encoding the Avro-formated alert.
        return_as: Format the alert will be returned in. One of 'dict' or 'df'.
        drop_cutouts: Whether to drop or return the cutouts (stamps).
                      If `return_as='df'` the cutouts are always dropped.
        kwargs: Keyword arguments passed to ``_drop_cutouts`` and
                ``alert_dict_to_dataframe``.

    Returns:
        alert packet in requested format
    """
    alert_dict = alert_avro_to_dict(alert_avro)

    if return_as == "dict":
        if drop_cutouts:
            return _drop_cutouts(alert_dict, **kwargs)
        else:
            return alert_dict

    elif return_as == "df":
        return alert_dict_to_dataframe(alert_dict, **kwargs)

    else:
        raise ValueError("`return_as` must be one of 'dict' or 'df'.")


def alert_avro_to_dict(alert_avro: Union[str, Path, bytes]) -> dict:
    """Load an alert Avro to a dictionary.

    Args:
        alert_avro:   Either the path of Avro file to load, or
            the bytes encoding the Avro-formated alert.

    Returns:
        alert as a dict
    """
    if isinstance(alert_avro, str) or isinstance(alert_avro, Path):
        with open(alert_avro, 'rb') as fin:
            alert_list = [r for r in fastavro.reader(fin)]  # list of dicts
    elif isinstance(alert_avro, bytes):
        try:
            with BytesIO(alert_avro) as fin:
                alert_list = [r for r in fastavro.reader(fin)]  # list of dicts
        except ValueError:
            try:
                # this function is mis-named because here we accept json encoding.
                # consider re-encoding alert packets using original Avro format
                # throughout broker to give users a consistent end-product.
                # then this except block will be unnecessary.
                alert_list = [json.loads(alert_avro.decode("UTF-8"))]  # list of dicts
            except ValueError:
                msg = (
                    "alert_avro is a bytes object, but does not seem to be either "
                    "json or Avro encoded. Cannot decode."
                )
                raise ValueError(msg)
    else:
        msg = (
            "Unable to open alert_avro. "
            "It must be either the path to a valid Avro file as a string, "
            "or bytes encoding an Avro-formated alert."
        )
        raise TypeError(msg)

    # we expect the list to contain exactly 1 entry
    if len(alert_list) == 0:
        raise ValueError('The alert Avro contains 0 valid entries.')
    if len(alert_list) > 1:
        raise ValueError('The alert Avro contains >1 entry.')

    alert_dict = alert_list[0]
    return alert_dict


def alert_dict_to_dataframe(alert_dict: dict, schema_map: dict) -> "pd.DataFrame":
    """ Packages an alert into a dataframe.
    Adapted from: https://github.com/ZwickyTransientFacility/ztf-avro-alert/blob/master/notebooks/Filtering_alerts.ipynb
    """
    import pandas as pd
    src_df = pd.DataFrame(alert_dict[schema_map['source']], index=[0])
    prvs_df = pd.DataFrame(alert_dict[schema_map['prvSources']])
    df = pd.concat([src_df, prvs_df], ignore_index=True)

    # attach some metadata. note this may not be preserved after all operations
    # https://stackoverflow.com/questions/14688306/adding-meta-information-metadata-to-pandas-dataframe
    # make sure this does not overwrite existing columns
    ids = AlertIds(schema_map, alert_dict=alert_dict)
    if "objectId" not in df.keys():
        df.objectId = ids.objectId
    if "sourceId" not in df.keys():
        df.sourceId = ids.sourceId

    return df


def _drop_cutouts(alert_dict: dict, schema_map: dict) -> dict:
    """Drop the cutouts from the alert dictionary."""
    cutouts = [
        schema_map['cutoutScience'],
        schema_map['cutoutTemplate'],
        schema_map['cutoutDifference']
    ]

    if schema_map['SURVEY'] == 'decat':
        alert_lite = {k: v for k, v in alert_dict.items()}
        for co in cutouts:
            alert_lite[schema_map['source']].pop(co, None)
            for psource in alert_lite[schema_map['prvSources']]:
                psource.pop(co, None)

    elif schema_map['SURVEY'] in ['ztf',  'elasticc']:
        alert_lite = {k: v for k, v in alert_dict.items() if k not in cutouts}

    return alert_lite



#!/usr/bin/env python3
# -*- coding: UTF-8 -*-
"""The ``math`` module contains mathematical transforms for survey data."""
from typing import Tuple

from astropy.time import Time
import numpy as np


def mag_to_flux(mag: float, zeropoint: float, magerr: float) -> Tuple[float, float]:
    """Convert an AB magnitude and its error to fluxes."""
    flux = 10 ** ((zeropoint - mag) / 2.5)
    fluxerr = flux * magerr * np.log(10 / 2.5)
    return flux, fluxerr


def jd_to_mjd(jd: float) -> float:
    """Convert Julian Date to modified Julian Date."""
    return Time(jd, format='jd').mjd
